import sys, os; sys.path.append(os.path.abspath(".."));
import src.qbandas as qb;
import pandas as pd


if __name__ == '__main__':

    qb.pull_schema("bs397hfuy")
    
    #df = pd.read_csv("data.csv")
    #qb.upload_records(df, "Example Table", drop=True)

    #qb.set_args("Example Table", "names")
    #qb.add_args("Example Table", "names", more_args="replacement!")
    # qb.upload_records(df, "Example Table", debug=True, drop=True)





"""
CLI for qBandas.
"""

import argparse
from textwrap import fill

from .config import field_types
from .schema import pull_schema

argParser = argparse.ArgumentParser(
    prog = 'qbandas',
    description = 'CLI for qBandas.'
)

argParser.add_argument("-t", "--col-types", 
    action='store_true',
    help='List the colmn types and their arguments.',
)

argParser.add_argument("--head",
    action='store_true',
    help='Create an empty header file in the current working directory.'
)

argParser.add_argument("-s", "--schema",
    help='Pull a schema from QuickBase.',
    metavar="DBID",
    nargs=1,
    type=str,
    dest="DBID"
)

args = argParser.parse_args()
    
if args.col_types:
    out = "data types:\n"
    for key in field_types:
        stub = f"  -{key}"
        gap = ' '*(16 - len(stub)) if len(stub) < 16 else '\n\t\t'
        out += stub + gap + fill(field_types[key], 80, subsequent_indent="\t\t") + '\n'
    print(out, end='')
    exit()

if args.head:
    with open('headers.json', 'w') as f:
        f.write('{\n\t"QB-Realm-Hostname": "{QB-Realm-Hostname}",\n\t"User-Agent": "{User-Agent}",\n\t"Authorization": "{Authorization}"\n}')

if args.DBID:
    pull_schema(*args.DBID)


"""
Methods that deal with resolving the structure of local and remote (QuickBase) tables.
"""

import json, requests, os

def pull_schema(DBID: str, **kwargs) -> None:
    """
    Download a local copy of a table's structure from a QuickBase application.

    This operation is authorized by `./headers.json`. The schema will be placed into `./schemas/` where other qBandas operations can use it. 

    Parameters
    ----------
    DBID : str
        The unique identifier of the table in QuickBase.
    """
    # Can set the directory with dir=<dir>
    dir = kwargs['dir'] if 'dir' in kwargs else os.getcwd()

    # resolve the headers
    with open(os.path.join(dir, 'headers.json'), 'r') as f:
            headers = json.load(f)

    # send the request to quickbase
    params = {
        'tableId': DBID,
        'includeFieldPerms': "false"
    }
    r = requests.get(
        'https://api.quickbase.com/v1/fields', 
        params = params, 
        headers = headers
    )
    r.raise_for_status()

    # convert the language in the api to user language
    type_conversion = {
        "timestamp": "datetime",
        "recordid": "numeric",
        "email": "email-address",
        "phone": "phone-number"
    }

    address_subfields = {
        ": Street 1": 1,
        ": Street 2": 2,
        ": City": 3,
        ": State": 4,
        ": Zip": 5,
        ": Country": 6
    }

    # parse the schema from the response
    schema = dict()
    schema['_DBID_'] = DBID
    for field in r.json():

        # address fields send extra unlabeled info that we cannot use. remove it
        if field['label'] in ['Street 1', 'Street 2', 'City', 'State/Region', 'Postal Code', 'Country']:
            continue

        _type = field['fieldType']
        if _type in type_conversion:
            _type = type_conversion[_type]
        
        schema[field['label']] = {
            'id': field['id'],
            'type': _type
        }

        # recreate subfields for adresses with propper names
        if _type == 'address':
            for sufix, fid_offset in address_subfields.items():
                schema[field['label'] + sufix] = {
                    'id': field['id'] + fid_offset,
                    'type': "text"
                }

    # dump the schmea to disk
    if not os.path.exists(os.path.join(dir, 'schemas')):
        os.makedirs(os.path.join(dir, 'schemas'))
    with open(os.path.join(dir, 'schemas', DBID + '.json'), 'w+') as f:
        json.dump(schema, f, indent=4)
    
        
def add_args(schema_name: str, field: str, **kwargs):
    """
    Append new config options (args) for a field in a schema.

    Parameters
    ----------
    schema_name : str
        The schema to modify.
    field : str
        This field will be configured.
    **kwargs
        The arguments to add.

    Examples
    --------
    ``` 
    >>> import qbandas as qb
    TODO
    ```
    """

    # Can set the directory with dir=<dir>
    dir = os.getcwd()
    if 'dir' in kwargs:
        dir = kwargs['dir']
        del kwargs['dir']

    # read in the schema
    file_name = schema_name + '.json'
    with open(os.path.join(dir, 'schemas', file_name), 'r') as f:
        schema = json.load(f)

    # append the new arguments
    if field not in schema:
        raise Exception(f"The field {field} is not in the schema {schema_name}")
    if 'args' not in schema[field]:
        schema[field]['args'] = kwargs
    else:
        schema[field]['args'] = schema[field]['args'] | kwargs

    # put the schema pack into the file
    with open(os.path.join(dir, 'schemas', file_name), 'w') as f:
        json.dump(schema, f, indent=4)

def set_args(schema_name: str, field: str, **kwargs):
    """
    Set the config options (args) for a field in a schema.

    Parameters
    ----------
    schema_name : str
        The schema to modify.
    field : str
        This field will be configured.
    **kwargs
        The arguments to add.
    """

    # Can set the directory with dir=<dir>
    dir = os.getcwd()
    if 'dir' in kwargs:
        dir = kwargs['dir']
        del kwargs['dir']

    # read in the schema
    file_name = schema_name + '.json'
    with open(os.path.join(dir, 'schemas', file_name), 'r') as f:
        schema = json.load(f)

    # set the arguments
    if field not in schema:
        raise Exception(f"The field {field} is not in the schema {schema_name}")
    schema[field]['args'] = kwargs

    # put the schema pack into the file
    with open(os.path.join(dir, 'schemas', file_name), 'w') as f:
        json.dump(schema, f, indent=4)


"""
Functions that deal with sending records or information to a Quickbase application.
"""

import requests, json, os
from functools import partial

import pandas as pd

from .config import field_types

def upload_records(df: pd.DataFrame, table_name: str, **kwargs): 
    """
    Send a table of records (rows) to a table on QuickBase.

    If no errors are thrown, the data was successfully uploaded to QuickBase. If an error occurs when sending the data, it is possible that only _some_ of your records will have been uploaded to QuickBase. If that is an issue, please fix the data, and send it all again.

    Parameters
    ----------
    df : pd.DataFrame
        The table of records
    table_name : str
        Identifies which table to send the data to. You should use `pull_schema()` to create a valid `table_name`.
    """

    debug = 'debug' in kwargs and kwargs['debug']
    if debug:
        from .util import str_dict, str_resp
        print(df.head())
    

    # setup: read in the headers and the schema
    # Can set the directory with dir=<dir>, defaults to .
    dir = kwargs['dir'] if 'dir' in kwargs else os.getcwd()
    with open(os.path.join(dir, 'headers.json'), 'r') as f:
            headers = json.load(f)
    with open(os.path.join(dir, 'schemas', table_name+'.json'), 'r') as f:
            schema = json.load(f)

    if debug:
        print(str_dict(headers))
        print(str_dict(schema))

    # Check that the columns in the df match the schema
    if 'drop' not in kwargs or not kwargs['drop']:
        unknown_columns = [] # columns in df that aren't in schema
        for column in df.columns:
            if column not in schema.keys():
                unknown_columns.append(column)
        if unknown_columns:
            raise Exception(f"The following columns from your dataframe have no matching column in the specified QuickBase table. If you want to drop these columns (not send their data), add `drop=True` to this function call. Columns: " + ', '.join(unknown_columns))
    
    # pack all the values into qb's crazy formats
    packed_df = pd.DataFrame()
    for col in df.columns:

        if col not in schema: # dropped columns
            continue

        col_info = schema[col]
        col_type = col_info['type']
        col_kwargs = col_info['args'] if 'args' in col_info else {}

        # Magic lines! Get the related parsing function from config, create a partial of it with the given column arguments, apply it to a copy of the current column, and save it in the output
        try:
            packing_func = field_types[col_type].packing_func
            packed_df[col] = df[col].copy().apply(partial(packing_func, **col_kwargs))
        except Exception as e:
            e.args = (f"Failed parsing column '{col}' with column type '{col_type}' and arguments '{col_kwargs}'", *e.args)
            raise

    if debug:
        print(packed_df.head())

    # grab the feild ids from schema, rename columns 
    fids = {k:v['id'] for k,v in schema.items() if k != '_DBID_' and k in packed_df.columns}
    renamed_df = packed_df.copy().rename(columns=fids) 

    # Convert each row to a dictonary, drop keys with null values, create the list of records
    def _row_to_dict(row):
        "Convert a row into a dictonary"
        row = dict(row)                                         
        return {k: v for k, v in row.items() if not pd.isna(v)} 
    records = list(renamed_df.apply(_row_to_dict, axis=1)) 

    if debug:
        print(str_dict(records[:5]))

     # setup destination information
    body = {"to": schema['_DBID_']}

    # look for optinal arguments to include from kwargs
    for option in ["fieldsToReturn", 'mergeFieldId']:
        if option in kwargs:
            body[option] = kwargs[option]

    # send the data one batch a time
    batch_size = kwargs["Payload-Size"] if "Payload-Size" in kwargs else 20_000
    for i in range(0, len(records), batch_size):
        body["data"] = records[i:i+batch_size]
        r = requests.post(
            'https://api.quickbase.com/v1/records', 
            headers = headers, 
            json = body
        )
        if debug:
            print(str_resp(r))
        r.raise_for_status()
    